
from .Marty import Marty
from .ClientMV1 import ClientMV1
from .ClientMV2 import ClientMV2
from .RICCommsSerial import RICCommsSerial
from .RICCommsWiFi import RICCommsWiFi
from .Exceptions import *

__version__ = '3.1.0a1'



import time
import logging
import os
import sys
import json
from martypy import Marty

jointNames = [
    'left hip',
    'left twist',
    'left knee',
    'right hip',
    'right twist',
    'right knee',
    'left arm',
    'right arm', 
    'eyes'   
]

def betweenCommands():
    time.sleep(3)

def testBoolCmd(cmdStr: str, cmdRslt: bool):
    logger.info(f"{cmdStr}, rslt = {cmdRslt}")
    betweenCommands()

def loggingCB(logStr: str) -> None:
    logger.debug(logStr)

# Check the log folder exists
logsFolder = "logs"
if not os.path.exists(logsFolder):
    os.mkdir(logsFolder)

# Log file name
logFileName = "MartyPyLogTest_" + time.strftime("%Y%m%d-%H%M%S") + ".log"
logFileName = os.path.join(logsFolder, logFileName)
print("Logging to file " + logFileName)

# Setup logging
logging.basicConfig(filename=logFileName, format='%(levelname)s: %(asctime)s %(funcName)s(%(lineno)d) -- %(message)s', level=logging.DEBUG)
logger = logging.getLogger("MartyPyTest")
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
logger.addHandler(handler)

mymarty = None
try:
    # mymarty = Marty('wifi', '192.168.86.11')
    # mymarty = Marty('wifi', '192.168.86.11', subscribeRateHz=0)
    mymarty = Marty('socket://192.168.86.41')
    # mymarty = Marty("usb", "COM9", debug=True)
    # mymarty = Marty('usb:///dev/tty.SLAB_USBtoUART', debug=True)
except Exception as excp:
    logger.debug(f"Couldn't connect to marty {excp}")
    exit()

mymarty.register_logging_callback(loggingCB)

martySysInfo = mymarty.get_system_info()
martyVersion2 = martySysInfo.get("HardwareVersion", "1.0") == "2.0"

if martyVersion2:
    logger.info(f"Marty has {len(mymarty.get_hw_elems_list())} hardware parts")

if martyVersion2:
    logger.info(f"Calibration flag {mymarty.is_calibrated()}")
testBoolCmd("Calibration flag clear", mymarty.clear_calibration())
if martyVersion2:
    logger.info(f"Calibration flag should be False ... {mymarty.is_calibrated()}")
    assert not mymarty.is_calibrated()
testBoolCmd("Calibration flag set", mymarty.save_calibration())
if martyVersion2:
    logger.info(f"Calibration flag should be True ... {mymarty.is_calibrated()}")
time.sleep(0.1)
if martyVersion2:
    assert mymarty.is_calibrated()

logger.info(f"Marty interface stats {json.dumps(mymarty.get_interface_stats())}")

testBoolCmd("Get ready", mymarty.get_ready())
testBoolCmd("Circle Dance", mymarty.circle_dance())
testBoolCmd("Eyes excited", mymarty.eyes('excited'))
testBoolCmd("Eyes wide", mymarty.eyes('wide'))
testBoolCmd("Eyes angry", mymarty.eyes('angry'))
testBoolCmd("Eyes normal", mymarty.eyes('normal'))
testBoolCmd("Kick left", mymarty.kick('left'))
testBoolCmd("Kick right", mymarty.kick('right'))
testBoolCmd("Stop", mymarty.stop())
testBoolCmd("Arms 45", mymarty.arms(45, 45, 500))
testBoolCmd("Arms 0", mymarty.arms(0, 0, 500))

testBoolCmd("Arms 0", mymarty.play_sound("disbelief"))
testBoolCmd("Arms 0", mymarty.play_sound("excited"))
testBoolCmd("Arms 0", mymarty.play_sound("screenfree"))

logger.info(f"Marty interface stats {json.dumps(mymarty.get_interface_stats())}")

for i in range(9): 
    testBoolCmd(f"Move joint {i}", mymarty.move_joint(i, i * 10, 500))
for jointName in jointNames: 
    testBoolCmd(f"Move joint {jointName}", mymarty.move_joint(jointName, 123, 500))

logger.info(f"Accelerometer x {mymarty.get_accelerometer('x')}")
logger.info(f"Accelerometer y { mymarty.get_accelerometer('y')}")
logger.info(f"Accelerometer z { mymarty.get_accelerometer('z')}")

if martyVersion2:
    testBoolCmd("Dance", mymarty.dance())
    testBoolCmd("Eyes wiggle", mymarty.eyes('wiggle'))
    testBoolCmd("Hold position", mymarty.hold_position(6000))
    testBoolCmd("is_moving", mymarty.is_moving())
    logger.info("Joint positions: ", [mymarty.get_joint_position(pos) for pos in range(9)])

time.sleep(5)

mymarty.close()



import logging
import os
import time
from typing import Callable, Dict, List, Optional, Union

from .ClientGeneric import ClientGeneric
from .RICCommsSerial import RICCommsSerial
from .RICCommsWiFi import RICCommsWiFi
from .RICCommsTest import RICCommsTest
from .RICProtocols import DecodedMsg, RICProtocols
from .RICROSSerial import RICROSSerial
from .RICInterface import RICInterface
from .RICHWElems import RICHWElems
from .Exceptions import (MartyConnectException,
                         MartyCommandException)

logger = logging.getLogger(__name__)

class ClientMV2(ClientGeneric):
    '''
    Lower level connector to Marty V2
    '''
    def __init__(self,
                method: str,
                locator: str,
                serialBaud: int = None,
                port = 80,
                wsPath = "/ws",
                subscribeRateHz = 10.0,
                *args, **kwargs):
        '''
        Initialise connection to remote Marty
        Args:
            client_type: 'wifi' (for WiFi), 'usb' (for usb serial), 'exp' (for expansion serial),
                'test' (output is available via get_test_output())
            locator: str, ipAddress, hostname, serial-port, name of test file etc
                    depending on method
            serialBaud: serial baud rate
            port: IP port for websockets
            wsPath: path to use for websocket connection
            subscribeRateHz: rate of fastest subscription to events
        Raises:
            MartyConnectException if the connection to the host failed
        '''
        # Call base constructor
        super().__init__(*args, **kwargs)

        # Initialise vars
        self.subscribeRateHz = subscribeRateHz
        self.lastSubscribedMsgTime = None
        self.maxTimeBetweenPubs = 10
        self.max_blocking_wait_time = 120  # seconds
        self.ricHardware = RICHWElems()
        self.isClosing = False
        self.ricSystemInfo = {}
        self.ricHwElemsInfoByIDNo = {}
        self.ricHwElemsList = []
        self.loggingCallback = None
        self._initComplete = False

        # Handle the method of connection
        if method == "usb" or method == "exp":
            ifType = "overascii" if method == "usb" else "plain"
            if serialBaud is None:
                serialBaud = 115200 if method == "usb" else 921600
            rifConfig = {
                "serialPort": locator,
                "serialBaud": serialBaud,
                "ifType": ifType,
            }
            self.ricIF = RICInterface(RICCommsSerial())
        elif method == "test":
            rifConfig = {
                "testFileName": locator
            }
            self.ricIF = RICInterface(RICCommsTest())
        else:
            rifConfig = {
                "ipAddrOrHostname": locator,
                "ipPort": port,
                "wsPath": wsPath,
                "ifType": "plain"
            }
            self.ricIF = RICInterface(RICCommsWiFi())

        # Open comms
        try:
            openOk = self.ricIF.open(rifConfig)
            if not openOk:
                raise MartyConnectException("Failed to open connection")
        except Exception as excp:
            raise MartyConnectException(str(excp))

        # Callbacks
        self.ricIF.setDecodedMsgCB(self._rxDecodedMsg)
        self.ricIF.setTimerCB(self._msgTimerCallback)
        self.ricIF.setLogLineCB(self._logDebugMsg)

    def start(self):
        self.ricSystemInfo = self.ricIF.cmdRICRESTSync("v")
        self._updateHwElemsInfo()
        self._initComplete = True

    def close(self):
        if self.isClosing:
            return
        self.isClosing = True
        # Stop any publishing messages
        self.ricIF.sendRICRESTCmdFrame('{"cmdName":"subscription","action":"update",' + \
                '"pubRecs":[' + \
                    '{"name":"MultiStatus","rateHz":0},' + \
                    '{"name":"PowerStatus","rateHz":0},' + \
                    '{"name":"AddOnStatus","rateHz":0}' + \
                ']}\0')
        # Allow message to be sent
        time.sleep(0.5)
        # Close the RIC interface
        self.ricIF.close()

    def wait_if_required(self, expected_wait_ms: int, blocking_override: Union[bool, None]):
        if not self.is_blocking(blocking_override):
            return

        deadline = time.time() + expected_wait_ms/1000 + self.max_blocking_wait_time
        time.sleep(2.5 * 1/self.subscribeRateHz)  # Give Marty time to report it is moving

        # The is_moving flag may be cleared briefly between 2 queued trajectories
        # Robot status may not be available for a while after Marty is turned on / connected to
        while self.is_moving() or self.get_robot_status().get('workQCount', 1) > 0:
            time.sleep(0.2 * 1/self.subscribeRateHz)
            if time.time() > deadline:
                raise TimeoutError("Marty wouldn't stop moving. Are you also controlling it via another method?"
                                   f"{os.linesep}If you issued some very long-running non-blocking commands, "
                                   "try increasing `marty.client.max_blocking_wait_time` (the current value "
                                   f"is {self.max_blocking_wait_time} s)")

    def hello(self) -> bool:
        return self.ricIF.cmdRICRESTRslt("traj/getReady")

    def get_ready(self) -> bool:
        return self.ricIF.cmdRICRESTRslt("traj/getReady")

    def stand_straight(self, move_time: int = 2000) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"traj/standStraight?moveTime={move_time}")

    def discover(self) -> List[str]:
        return []

    def stop(self, stop_type: str, stopCode: int) -> bool:
        robotRestCmd = "stop"
        trajCmd = ""
        if stopCode == 0: robotRestCmd = "stopAfterMove"
        elif stopCode == 2: robotRestCmd = "panic"
        elif stopCode == 3: trajCmd = "getReady"
        elif stopCode == 4 or stopCode == 5: robotRestCmd = "pause"
        isOk = self.ricIF.cmdRICRESTRslt("robot/" + robotRestCmd)
        if len(trajCmd) > 0:
            self.ricIF.cmdRICRESTRslt("traj/" + trajCmd)
        return isOk

    def resume(self) -> bool:
        return self.ricIF.cmdRICRESTRslt("robot/resume")

    def hold_position(self, hold_time: int) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"traj/hold?moveTime={hold_time}")

    def move_joint(self, joint_id: int, position: int, move_time: int) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"traj/joint?jointID={joint_id}&angle={position}&moveTime={move_time}")

    def get_joint_position(self, joint_id: Union[int, str]) -> float:
        return self.ricHardware.getServoPos(joint_id, self.ricHwElemsInfoByIDNo)

    def get_joint_current(self, joint_id: Union[int, str]) -> float:
        return self.ricHardware.getServoCurrent(joint_id, self.ricHwElemsInfoByIDNo)

    def get_joint_status(self, joint_id: Union[int, str]) -> int:
        return self.ricHardware.getServoFlags(joint_id, self.ricHwElemsInfoByIDNo)

    def lean(self, direction: str, amount: Optional[int], move_time: int) -> bool:
        if amount is None:
            amount = 29
        try:
            directionNum = ClientGeneric.SIDE_CODES[direction]
        except KeyError:
            self.preException(True)
            raise MartyCommandException("Direction must be one of {}, not '{}'"
                                        "".format(set(ClientGeneric.SIDE_CODES.keys()), direction))
        return self.ricIF.cmdRICRESTRslt(f"traj/lean?side={directionNum}&leanAngle={amount}&moveTime={move_time}")

    def walk(self, num_steps: int = 2, start_foot:str = 'auto', turn: int = 0,
                step_length:int = 25, move_time: int = 1500) -> bool:
        side_url_param = ''
        if start_foot != 'auto':
            try:
                sideNum = ClientGeneric.SIDE_CODES[start_foot]
            except KeyError:
                raise MartyCommandException("Direction must be one of {}, not '{}'"
                                            "".format(set(ClientGeneric.SIDE_CODES.keys()), start_foot))
            side_url_param = f'&side={sideNum}'
        return self.ricIF.cmdRICRESTRslt(f"traj/step/{num_steps}?stepLength={step_length}&turn={turn}"
                                         f"&moveTime={move_time}" + side_url_param)

    def eyes(self, joint_id: int, pose_or_angle: Union[str, int], move_time: int = 100) -> bool:
        if type(pose_or_angle) is str:
            try:
                eyesTrajectory = ClientGeneric.EYE_POSES[pose_or_angle]
            except KeyError:
                raise MartyCommandException("pose must be one of {}, not '{}'"
                                            "".format(set(ClientGeneric.EYE_POSES.keys()), pose_or_angle))
            return self.ricIF.cmdRICRESTRslt(f"traj/{eyesTrajectory}")
        return self.move_joint(joint_id, pose_or_angle, move_time)

    def kick(self, side: str = 'right', twist: int = 0, move_time: int = 2500) -> bool:
        if side != 'right' and side != 'left':
            raise MartyCommandException("side must be one of 'right' or 'left', not '{}'"
                                        "".format(side))
        return self.ricIF.cmdRICRESTRslt(f"traj/kick?side={ClientGeneric.SIDE_CODES[side]}&moveTime={move_time}&turn={twist}")

    def arms(self, left_angle: int, right_angle: int, move_time: int) -> bool:
        self.move_joint(6, left_angle, move_time)
        return self.move_joint(7, right_angle, move_time)

    def celebrate(self, move_time: int = 4000) -> bool:

        # TODO - add celebrate trajectory to Marty V2
        return self.ricIF.cmdRICRESTRslt(f"traj/wiggle?moveTime={move_time}")

    def circle_dance(self, side: str = 'right', move_time: int = 2500) -> bool:
        if side != 'right' and side != 'left':
            raise MartyCommandException("side must be one of 'right' or 'left', not '{}'"
                                        "".format(side))
        return self.ricIF.cmdRICRESTRslt(f"traj/circle?side={ClientGeneric.SIDE_CODES[side]}&moveTime={move_time}")

    def dance(self, side: str = 'right', move_time: int = 4500) -> bool:
        if side != 'right' and side != 'left':
            raise MartyCommandException("side must be one of 'right' or 'left', not '{}'"
                                        "".format(side))
        return self.ricIF.cmdRICRESTRslt(f"traj/dance?side={ClientGeneric.SIDE_CODES[side]}&moveTime={move_time}")

    def wiggle(self, move_time: int = 4000) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"traj/wiggle?moveTime={move_time}")

    def sidestep(self, side: str, steps: int = 1, step_length: int = 35,
            move_time: int = 1000) -> bool:
        if side != 'right' and side != 'left':
            raise MartyCommandException("side must be one of 'right' or 'left', not '{}'"
                                        "".format(side))
        return self.ricIF.cmdRICRESTRslt(f"traj/sidestep/{steps}?side={ClientGeneric.SIDE_CODES[side]}"
                                         f"&stepLength={step_length}&moveTime={move_time}")

    def play_sound(self, name_or_freq_start: Union[str,float],
            freq_end: Optional[float] = None,
            duration: Optional[int] = None) -> bool:
        if not name_or_freq_start.lower().endswith(".raw"):
                name_or_freq_start += ".raw"
        return self.ricIF.cmdRICRESTRslt(f"filerun/{name_or_freq_start}")

    def pinmode_gpio(self, gpio: int, mode: str) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def write_gpio(self, gpio: int, value: int) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def digitalread_gpio(self, gpio: int) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def i2c_write(self, *byte_array: int) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def i2c_write_to_ric(self, address: int, byte_array: bytes) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_battery_voltage(self) -> float:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_battery_remaining(self) -> float:
        powerStatus = self.ricHardware.getPowerStatus()
        return powerStatus.get("remCapPC", 0)

    def get_distance_sensor(self) -> float:
        # TODO NotImplemented
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_accelerometer(self, axis: Optional[str] = None, axisCode: int = 0) -> float:
        if axis is None:
            return self.ricHardware.getIMUAll()
        return self.ricHardware.getIMUAxisValue(axisCode)

    def enable_motors(self, enable: bool = True, clear_queue: bool = True) -> bool:
        return True

    def enable_safeties(self, enable: bool = True) -> bool:
        return True

    def fall_protection(self, enable: bool = True) -> bool:
        return True

    def motor_protection(self, enable: bool = True) -> bool:
        return True

    def battery_protection(self, enable: bool = True) -> bool:
        return True

    def buzz_prevention(self, enable: bool = True) -> bool:
        return True

    def lifelike_behaviour(self, enable: bool = True) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def set_parameter(self, *byte_array: int) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def save_calibration(self) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"calibrate/set")

    def clear_calibration(self) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"calibrate/setFlag/0")

    def is_calibrated(self) -> bool:
        result = self.ricIF.cmdRICRESTSync("calibrate")
        if result.get("rslt", "") == "ok":
            return result.get("calDone", 0) != 0
        return False

    def ros_command(self, *byte_array: int) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def keyframe (self, time: float, num_of_msgs: int, msgs) -> List[bytes]:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_chatter(self) -> bytes:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_firmware_version(self) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def _mute_serial(self) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def ros_serial_formatter(self, topicID: int, send: bool = False, *message: int) -> List[int]:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def is_moving(self) -> bool:
        return self.ricHardware.getIsMoving()

    def is_paused(self) -> bool:
        return self.ricHardware.getIsPaused()

    def get_robot_status(self) -> Dict:
        return self.ricHardware.getRobotStatus()

    def get_joints(self) -> Dict:
        return self.ricHardware.getServos(self.ricHwElemsInfoByIDNo)

    def get_power_status(self) -> Dict:
        return self.ricHardware.getPowerStatus()

    def get_add_ons_status(self) -> Dict:
        return self.ricHardware.getAddOns(self.ricHwElemsInfoByIDNo)

    def get_add_on_status(self, add_on_name_or_id: Union[int, str]) -> Dict:
        return self.ricHardware.getAddOn(add_on_name_or_id, self.ricHwElemsInfoByIDNo)

    def get_system_info(self) -> Dict:
        return self.ricSystemInfo

    def set_marty_name(self, name: str) -> bool:
        escapedName = name.replace('"', '').replace('\n','')
        return self.ricIF.cmdRICRESTRslt(f"friendlyname/{escapedName}")

    def get_marty_name(self) -> str:
        result = self.ricIF.cmdRICRESTSync("friendlyname")
        if result.get("rslt", "") == "ok":
            return result.get("friendlyName", "Marty")
        return "Marty"

    def is_marty_name_set(self) -> bool:
        result = self.ricIF.cmdRICRESTSync("friendlyname")
        if result.get("rslt", "") == "ok":
            return result.get("friendlyNameIsSet", 0) != 0
        return False

    def get_hw_elems_list(self) -> List:
        self._updateHwElemsInfo()
        return self.ricHwElemsList

    def send_ric_rest_cmd(self, ricRestCmd: str) -> None:
        self.ricIF.sendRICRESTURL(ricRestCmd)

    def send_ric_rest_cmd_sync(self, ricRestCmd: str) -> Dict:
        return self.ricIF.cmdRICRESTSync(ricRestCmd)

    def register_logging_callback(self, loggingCallback: Callable[[str],None]) -> None:
        self.loggingCallback = loggingCallback

    def get_interface_stats(self) -> Dict:
        return self.ricIF.getStats()

    def preException(self, isFatal: bool) -> None:
        if isFatal:
            self.ricIF.close()
        logger.debug(f"Pre-exception isFatal {isFatal}")

    def _rxDecodedMsg(self, decodedMsg: DecodedMsg, interface: RICInterface):
        if decodedMsg.protocolID == RICProtocols.PROTOCOL_ROSSERIAL:
            # logger.debug(f"ROSSERIAL message received {len(decodedMsg.payload)}")
            self.lastSubscribedMsgTime = time.time()
            if decodedMsg.payload:
                RICROSSerial.decode(decodedMsg.payload, 0, self.ricHardware.updateWithROSSerialMsg)
        elif decodedMsg.protocolID == RICProtocols.PROTOCOL_RICREST:
            # logger.debug(f"RIC REST message received {decodedMsg.payload}")
            pass

    def _logDebugMsg(self, logMsg: str) -> None:
        if self.loggingCallback:
            self.loggingCallback(logMsg)

    def _msgTimerCallback(self) -> None:
        if self.isClosing:
            return
        if (self._initComplete and self.subscribeRateHz != 0) and \
                 (self.lastSubscribedMsgTime is None or \
                    time.time() > self.lastSubscribedMsgTime + self.maxTimeBetweenPubs):
            # Subscribe for publication messages
            self.ricIF.sendRICRESTCmdFrame('{"cmdName":"subscription","action":"update",' + \
                            '"pubRecs":[' + \
         '{' + f'"name":"MultiStatus","rateHz":{self.subscribeRateHz},' + '}' + \
                                '{"name":"PowerStatus","rateHz":1.0},' + \
                                '{' + f'"name":"AddOnStatus","rateHz":{self.subscribeRateHz}' + '}' + \
                            ']}\0')
            self.lastSubscribedMsgTime = time.time()

    def _updateHwElemsInfo(self):
        hwElemsInfo = self.ricIF.cmdRICRESTSync("hwstatus")
        if hwElemsInfo.get("rslt", "") == "ok":
            self.ricHwElemsList = hwElemsInfo.get("hw", [])
            self.ricHwElemsInfoByIDNo = {}
            for el in self.ricHwElemsList:
                if "IDNo" in el:
                    self.ricHwElemsInfoByIDNo[el["IDNo"]] = el

    def get_test_output(self) -> dict:
        return self.ricIF.getTestOutput()



'''
Serial communications with a Robotical RIC
'''
from threading import Thread
from typing import Callable, Dict, List, Union
import serial
import serial.tools.list_ports
import time
import logging
from serial.serialutil import SerialException
from .RICCommsBase import RICCommsBase
from .LikeHDLC import LikeHDLC
from .ProtocolOverAscii import ProtocolOverAscii
from .Exceptions import MartyConnectException

logger = logging.getLogger(__name__)

class RICCommsSerial(RICCommsBase):
    '''
    RICCommsSerial
    Provides a serial interface for RIC communications
    '''
    def __init__(self) -> None:
        '''
        Initialise RICCommsSerial
        '''
        super().__init__()
        self._isOpen = False
        self.serialReaderThread: Thread = None
        self.serialDevice: serial.Serial = None
        self.serialThreadEnabled = False
        self._hdlc = LikeHDLC(self._onHDLCFrame, self._onHDLCError)
        self.overAscii = False
        self.serialLogLine = ""
        self.serialPortErrors = 0

    def __del__(self) -> None:
        '''
        Destructor
        '''
        self.close()

    @classmethod
    def detect_rics(cls) -> List[str]:
        serial_ports = serial.tools.list_ports.comports()
        possible_rics = [port.device for port in serial_ports
                         if port.description.startswith("CP2102")]
        return possible_rics

    def isOpen(self) -> bool:
        '''
        Check if comms open
        Returns:
            True if comms open
        '''
        return self._isOpen

    def open(self, openParams: Dict) -> bool:
        '''
        Open connection
        Protocol can be plain (if the port is not used for any other purpose) or
                 overascii (if the port is also used for logging information)
        Args:
            openParams: dict containing params used to open the connection, may include
                        "serialPort" (will be auto-detected if missing/empty),
                        "serialBaud",
                        "ifType" == "plain" or "overascii"
        Returns:
            True if open succeeded or already open
        Throws:
            MartyConnectException: if connection cannot be opened
        '''
        # Check not already open
        if self._isOpen:
            return True

        # Get params
        serialPort = openParams.get("serialPort", "")
        serialBaud = openParams.get("serialBaud", 115200)
        self.overAscii = openParams.get("ifType", "plain") != "plain"
        if self.overAscii:
            self.protocolOverAscii = ProtocolOverAscii()

        # Open serial port
        rics = self.detect_rics()
        try:
            self.serialDevice = serial.Serial(port=None, baudrate=serialBaud)
            self.serialDevice.port = serialPort or rics[0]
            self.serialDevice.rts = 0
            self.serialDevice.dtr = 0
            self.serialDevice.dsrdtr = False
            self.serialDevice.open()
        except Exception as excp:
            error_message = f"Serial port problem on {self.serialDevice.port}. Try "\
                            f"connecting to one of the following ports instead: {rics}"
            raise MartyConnectException(error_message) from excp

        # Start receive loop
        self.serialThreadEnabled = True
        self.serialReaderThread = Thread(target=self._serialRxLoop)
        self.serialReaderThread.daemon = True
        self.serialReaderThread.start()
        self._isOpen = True
        return True

    def close(self) -> None:
        '''
        Close serial port
        '''
        if not self._isOpen:
            return
        # Stop thread function
        if self.serialReaderThread is not None:
            self.serialThreadEnabled = False
            time.sleep(0.01)
            self.serialReaderThread.join()
            self.serialReaderThread = None
        # Close port
        if self.serialDevice is not None:
            self.serialDevice.close()
            self.serialDevice = None
        self._isOpen = False

    def send(self, data: bytes) -> None:
        '''
        Send data to serial port
        Args:
            data: bytes to send on serial port
        Returns:
            none
        Throws:
            MartyConnectException: if the connection has an error
        '''
        # logger.debug(f"Sending to IF len {len(bytesToSend)} {str(bytesToSend)}")
        hdlcEncoded = self._hdlc.encode(data)
        try:
            if self.overAscii:
                encodedFrame = ProtocolOverAscii.encode(hdlcEncoded)
                self._sendBytesToIF(encodedFrame)
                # logger.debug(f"send {''.join('{:02x}'.format(x) for x in encodedFrame)}")
            else:
                self._sendBytesToIF(hdlcEncoded)
        except Exception as excp:
            raise MartyConnectException("Serial send problem") from excp

    def _serialRxLoop(self) -> None:
        '''
        Thread function used to process serial port data
        '''
        while self.serialThreadEnabled:
            i = self.serialDevice.in_waiting
            if i < 1:
                time.sleep(0.001)
                continue
            byt = self.serialDevice.read(i)
            for b in byt:
                if self.overAscii:
                    if b >= 128:
                        decodedVal = self.protocolOverAscii.decodeByte(b)
                        if decodedVal >= 0:
                            # logger.debug(f"{decodedVal:02x}")
                            self._hdlc.decodeData(decodedVal)
                    else:
                        if b == 0x0a:
                            if self.logLineCB:
                                self.logLineCB(self.serialLogLine)
                            self.serialLogLine = ""
                        else:
                            self.serialLogLine += chr(b)
                else:
                    self._hdlc.decodeData(b)
            # logger.debug(f"CommsSerial rx {''.join('{:02x}'.format(x) for x in byt)}")
        # logger.debug("Exiting serialRxLoop")

    def _onHDLCFrame(self, frame: bytes) -> None:
        if self.rxFrameCB is not None:
            self.rxFrameCB(frame)

    def _onHDLCError(self) -> None:
        pass

    def _sendBytesToIF(self, bytesToSend: bytes) -> None:
        # logger.debug(f"Sending to IF len {len(bytesToSend)} {str(bytesToSend)}")
        if not self._isOpen:
            return
        # logger.debug(f"CommsSerial sendBytesToIF {''.join('{:02x}'.format(x) for x in bytesToSend)}")
        if self.serialDevice is not None:
            try:
                self.serialDevice.write(bytesToSend)
            except SerialException as excp:
                # The port has connected but now failing to send characters
                # This may sort itself out or require user intervention
                self.serialPortErrors += 1
                pass

    def getTestOutput(self) -> dict:
        return {}



'''
RICHWElems
Class supporting the hardware on RIC
'''

import logging
import time
from martypy.RICROSSerial import RICROSSerial
from typing import Dict, List, Tuple, Union

logger = logging.getLogger(__name__)

class RICHwSmartServos:
    def __init__(self) -> None:
        self.latestMsg: bytes = None
        self.latestMsgTime: float = None

    def update(self, msgPayload: bytes) -> None:
        self.latestMsg = msgPayload
        self.latestMsgTime = time.time()

    def status(self, dictOfHwElemsByIdNo: Dict) -> Dict:
        if self.latestMsgTime is None:
            return {}
        servosStatus = RICROSSerial.extractSmartServos(self.latestMsg)
        fieldsToCopy = ["name"]
        for IDNo in servosStatus:
            if IDNo in dictOfHwElemsByIdNo:
                for field in fieldsToCopy:
                    if field in dictOfHwElemsByIdNo[IDNo]:
                        servosStatus[IDNo][field] = dictOfHwElemsByIdNo[IDNo][field]
        return servosStatus

    def servoStatus(self, servoId: int, dictOfHwElemsByIdNo: Dict) -> Dict:
        status = self.status(dictOfHwElemsByIdNo)
        return status.get(servoId, {})

class RICHwIMU:
    def __init__(self) -> None:
        self.latestMsg: bytes = None
        self.latestMsgTime: float = None

    def update(self, msgPayload: bytes) -> None:
        if len(msgPayload) != RICROSSerial.ROS_ACCEL_BYTES:
            return
        self.latestMsg = msgPayload
        self.latestMsgTime = time.time()
        # logger.debug(f"IMU update len {len(msgPayload)}")

    def xyz(self) -> Tuple[float, float, float]:
        if self.latestMsgTime is None:
            return (0,0,0)
        return RICROSSerial.extractAccel(self.latestMsg)

    def axisVal(self, axisCode: int) -> float:
        if self.latestMsgTime is None:
            return 0
        xyzTuple = RICROSSerial.extractAccel(self.latestMsg)
        return xyzTuple[axisCode]

class RICHwPowerStatus:
    def __init__(self) -> None:
        self.latestMsg: bytes = None
        self.latestMsgTime: float = None

    def update(self, msgPayload: bytes) -> None:
        if len(msgPayload) != RICROSSerial.ROS_POWER_STATUS_BYTES:
            return
        self.latestMsg = msgPayload
        self.latestMsgTime = time.time()

    def powerStatus(self) -> Dict:
        if self.latestMsgTime is None:
            return {}
        return RICROSSerial.extractPowerStatus(self.latestMsg)

class RICHwAddOnStatus:
    def __init__(self):
        self.latestMsg: bytes = None
        self.latestMsgTime: float = None
        self.addOnNameToIdMap = {}

    def update(self, msgPayload: bytes) -> None:
        self.latestMsg = msgPayload
        self.latestMsgTime = time.time()

    def status(self, dictOfHwElemsByIdNo: Dict) -> Dict:
        if self.latestMsgTime is None:
            return {}
        addOnStatus = RICROSSerial.extractAddOnStatus(self.latestMsg)
        fieldsToCopy = ["name", "type", "whoAmITypeCode"]
        for IDNo in addOnStatus:
            if IDNo in dictOfHwElemsByIdNo:
                for field in fieldsToCopy:
                    if field in dictOfHwElemsByIdNo[IDNo]:
                        addOnStatus[IDNo][field] = dictOfHwElemsByIdNo[IDNo][field]
        return addOnStatus

    def addOnStatus(self, addOnNameOrId: Union[int, str], dictOfHwElemsByIdNo: Dict) -> Dict:
        status = self.status(dictOfHwElemsByIdNo)
        addOnId = 0
        if type(addOnNameOrId) is str:
            if addOnNameOrId in self.addOnNameToIdMap:
                addOnId = self.addOnNameToIdMap.get(addOnNameOrId, 0)
        else:
            addOnId = addOnNameOrId
        return status.get(addOnId, {})

class RICHwRobotStatus:
    def __init__(self):
        self.latestMsg: bytes = None
        self.latestMsgTime: float = None

    def update(self, msgPayload: bytes) -> None:
        if len(msgPayload) != RICROSSerial.ROS_ROBOT_STATUS_BYTES:
            return
        self.latestMsg = msgPayload
        self.latestMsgTime = time.time()

    def status(self) -> Dict:
        if self.latestMsgTime is None:
            return {}
        return RICROSSerial.extractRobotStatus(self.latestMsg)

class RICHWElems:

    def __init__(self):
        self._smartServos = RICHwSmartServos()
        self._IMU = RICHwIMU()
        self._powerStatus = RICHwPowerStatus()
        self._addOnsStatus = RICHwAddOnStatus()
        self._robotStatus = RICHwRobotStatus()

    def updateWithROSSerialMsg(self, topicID: int, payload: bytes):
        # logger.debug(f"Received ROSSerial topicID {topicID}")
        if topicID == RICROSSerial.ROSTOPIC_V2_SMART_SERVOS:
            self._smartServos.update(payload)
        elif topicID == RICROSSerial.ROSTOPIC_V2_ACCEL:
            self._IMU.update(payload)
        elif topicID == RICROSSerial.ROSTOPIC_V2_POWER_STATUS:
            self._powerStatus.update(payload)
        elif topicID == RICROSSerial.ROSTOPIC_V2_ADDONS:
            self._addOnsStatus.update(payload)
        elif topicID == RICROSSerial.ROSTOPIC_V2_ROBOT_STATUS:
            self._robotStatus.update(payload)

    def getServos(self, dictOfHwElemsByIdNo: Dict) -> List:
        return self._smartServos.status(dictOfHwElemsByIdNo)

    def getServoPos(self, servoId: int, dictOfHwElemsByIdNo: Dict) -> float:
        return self._smartServos.servoStatus(servoId, dictOfHwElemsByIdNo).get("pos", 0)

    def getServoCurrent(self, servoId: int, dictOfHwElemsByIdNo: Dict) -> float:
        return self._smartServos.servoStatus(servoId, dictOfHwElemsByIdNo).get("current", 0)

    def getServoFlags(self, servoId: int, dictOfHwElemsByIdNo: Dict) -> int:
        return self._smartServos.servoStatus(servoId, dictOfHwElemsByIdNo).get("flags", 0)

    def getIMUAll(self) -> Tuple[float, float, float]:
        return self._IMU.xyz()

    def getIMUAxisValue(self, axisCode: int) -> float:
        return self._IMU.axisVal(axisCode)

    def getPowerStatus(self) -> Dict:
        return self._powerStatus.powerStatus()

    def getRobotStatus(self) -> Dict:
        return self._robotStatus.status()

    def getIsMoving(self) -> bool:
        return self._robotStatus.status().get("isMoving", False)

    def getIsPaused(self) -> bool:
        return self._robotStatus.status().get("isPaused", False)

    def getAddOns(self, dictOfHwElemsByIdNo: Dict) -> List:
        return self._addOnsStatus.status(dictOfHwElemsByIdNo)

    def getAddOn(self, addOnNameOrId: Union[int, str], dictOfHwElemsByIdNo: Dict) -> Dict:
        return self._addOnsStatus.addOnStatus(addOnNameOrId, dictOfHwElemsByIdNo)

